/*
* Author: Chris Campbell - www.iforce2d.net
*
* See usage details at http://www.iforce2d.net/cc2dxres/otf
*
* This software is provided 'as-is', without any express or implied warranty.
* In no event will the authors be held liable for any damages arising from
* the use of this software. Permission is granted to anyone to use this software
* for any purpose, including commercial applications, and to alter it and
* redistribute it freely.
*
*/
#include "OTFLayer.h"
#include "ScreenLog.h"
#include "md5.h"
#include "json/json.h"
#include "iforce2dFileUtils.h"

#include "GameScreen.h" // change this to the layer your app would normally start with

using namespace std;
using namespace cocos2d;
using namespace cocos2d::network;

string otfSourceURL     = "http://www.iforce2d.net/otfdemo/";
string otfLocalFolder   = "otf";
string otfManifestFile  = "otf-manifest.php";

OTFLayer::OTFLayer()
{
    m_otfUpdateInProgress = false;
    m_requestInProgress = false;
    m_otfSkippedCount = 0;
    m_otfFetchedCount = 0;
    m_md5Buffer = new unsigned char[128 * 1024];
}

OTFLayer::~OTFLayer()
{
    unscheduleUpdate();
    delete[] m_md5Buffer;
}

CCScene* OTFLayer::scene()
{
    auto scene = CCScene::create();
    auto layer = OTFLayer::create();
    scene->addChild(layer);

    return scene;
}

bool OTFLayer::init()
{
    CCLayer::init();
    setTouchEnabled( true );
    scheduleUpdate();
    updateOTFContent();
    return true;
}

void OTFLayer::update(float dt)
{
    if ( m_otfUpdateInProgress )
        otfTick();

    if ( g_screenLog )
        g_screenLog->update(dt);
}

void OTFLayer::onTouchesBegan(const vector<cocos2d::Touch*>& touches, Event *event)
{
    if ( m_otfUpdateInProgress )
        return;

    auto scene = GameScreen::scene(); // change this to the layer your app would normally start with
    g_screenLog->attachToScene(scene);
    CCDirector::sharedDirector()->replaceScene( scene );
}

void OTFLayer::startHttpGET(string url)
{
    m_requestInProgress = true;

    string fullURL = otfSourceURL + url;
    g_screenLog->log(LL_INFO, "Fetching %s...", fullURL.c_str());
    auto request = new HttpRequest();
    request->setUrl( fullURL.c_str() );
    request->setRequestType(HttpRequest::Type::GET);
    request->setResponseCallback( this, httpresponse_selector(OTFLayer::onRequestCompleted));
    request->setTag( url.c_str() );
    HttpClient::getInstance()->send(request);
}

string OTFLayer::getHashForFile( string filename )
{
    FILE* f = fopen(filename.c_str(), "rb");
    if ( ! f ) {
        return "";
    }

    int bufferSize = 128 * 1024;

    unsigned char digest[16];
    MD5_CTX context;
    MD5_Init(&context);

    bool done = false;
    int pos = 0;
    while ( ! done ) {
        size_t bytesRead = fread( m_md5Buffer + pos, 1, bufferSize, f );
        if ( bytesRead == 0 )
            done = true;
        else
            MD5_Update(&context, m_md5Buffer, bytesRead);
    }

    fclose(f);

    MD5_Final(digest, &context);

    char md5string[33];
    md5string[32] = 0;
    for(int i = 0; i < 16; ++i)
        sprintf(&md5string[i*2], "%02x", (unsigned int)digest[i]);

    return md5string;
}

void OTFLayer::handleManifest(cocos2d::network::HttpResponse *response )
{
    string data( response->getResponseData()->begin(), response->getResponseData()->end() );

    if ( response->getResponseCode() != 200 ) {
        g_screenLog->log(LL_ERROR, "Could not download manifest file - is there an active internet connection available?");
        return;
    }

    Json::Value manifestValue;
    Json::Reader reader;
    bool parsed = reader.parse(data, manifestValue);
    if ( ! parsed ) {
        g_screenLog->log(LL_ERROR, "Failed to parse manifest : %s", reader.getFormatedErrorMessages().c_str());
        return;
    }

    if ( ! manifestValue.isMember("entries")  ) {
        g_screenLog->log(LL_ERROR, "Could not find 'entries' in manifest.");
        return;
    }

    Json::Value entries = manifestValue["entries"];
    g_screenLog->log(LL_INFO, "Manifest contains %d entries.", entries.size());

    FileUtils* pSharedFileUtils = FileUtils::sharedFileUtils();

    for ( int i = 0; i < entries.size(); i++ ) {
        Json::Value entry = entries[i];
        string file = entry.get("f", "").asString();
        string hash = entry.get("h", "").asString();

        string existingHash = getHashForFile( pSharedFileUtils->getWritablePath() + otfLocalFolder + "/" + file );

        if ( existingHash == hash ) {
            CCLOG("Skipping: %s", file.c_str());
            m_otfSkippedCount++;
        }
        else {
            m_urlsToCheck.push_back( make_pair(file,hash) );
        }
    }

    g_screenLog->log(LL_INFO, "Found %d files needing update.", m_urlsToCheck.size());
}

void OTFLayer::handleResourceFile( HttpResponse* response )
{    
    string filename = otfLocalFolder + "/" + response->getHttpRequest()->getTag();

    FileUtils* pSharedFileUtils = FileUtils::sharedFileUtils();
    string expectedPath = pSharedFileUtils->getWritablePath() + filename;

    size_t posOfLastSlash = filename.find_last_of('/');
    if ( posOfLastSlash != string::npos ) {
        string folder = filename.substr(0, posOfLastSlash);
        createFolder( pSharedFileUtils->getWritablePath() + folder );
    }

    FILE* f = fopen(expectedPath.c_str(), "wb");
    if ( f ) {
        std::vector<char> *buffer = response->getResponseData();
        int wroteBytes = fwrite( &(*buffer)[0], 1, buffer->size(), f );
        if ( wroteBytes != buffer->size() )
            g_screenLog->log(LL_ERROR, "Failed to write file %s", expectedPath.c_str());
        fclose(f);

        m_otfFetchedCount++;
    }
    else {
        g_screenLog->log(LL_ERROR, "Could not open file for writing: %s", expectedPath.c_str());
    }
}

void OTFLayer::onRequestCompleted(HttpClient* client, HttpResponse* response)
{
    if ( response->getHttpRequest()->getTag() == otfManifestFile )
        handleManifest(response);
    else
        handleResourceFile(response);

    m_requestInProgress = false;
}

void OTFLayer::updateOTFContent()
{
    // override the path used to find files
    FileUtils* pFileUtils = FileUtils::sharedFileUtils();
    vector<string> paths = { pFileUtils->getWritablePath() + otfLocalFolder };
    pFileUtils->setSearchPaths( paths );

    g_screenLog->log(LL_INFO, "Starting OTF update...");
    m_otfSkippedCount = 0;
    m_otfFetchedCount = 0;

    m_otfUpdateInProgress = true;
    startHttpGET( otfManifestFile );
}

void OTFLayer::otfTick()
{
    if ( m_requestInProgress )
        return;

    if ( ! m_urlsToCheck.empty() ) {
        m_currentUrl = m_urlsToCheck[0];
        m_urlsToCheck.erase( m_urlsToCheck.begin() );
        startHttpGET( m_currentUrl.first );
        return;
    }

    g_screenLog->log(LL_INFO, "Fetched %d files, skipped %d files.", m_otfFetchedCount, m_otfSkippedCount);
    g_screenLog->log(LL_INFO, "OTF update finished - tap the screen to continue.");

    VolatileTextureMgr::reloadAllTextures();

    m_otfUpdateInProgress = false;
}

